# distutils: language = c++


cdef extern from "HTS_engine.h":
    cdef cppclass _HTS_Engine:
        pass
    ctypedef _HTS_Engine HTS_Engine

    void HTS_Engine_initialize(HTS_Engine * engine)
    char HTS_Engine_load(HTS_Engine * engine, char **voices, size_t num_voices)
    size_t HTS_Engine_get_sampling_frequency(HTS_Engine * engine)
    size_t HTS_Engine_get_fperiod(HTS_Engine * engine)
    void HTS_Engine_refresh(HTS_Engine * engine)
    void HTS_Engine_clear(HTS_Engine * engine)
    const char *HTS_Engine_get_fullcontext_label_format(HTS_Engine * engine)
    char HTS_Engine_synthesize_from_strings(HTS_Engine * engine, char **lines, size_t num_lines)
    char HTS_Engine_synthesize_from_fn(HTS_Engine * engine, const char *fn)
    double HTS_Engine_get_generated_speech(HTS_Engine * engine, size_t index)
    size_t HTS_Engine_get_nsamples(HTS_Engine * engine)

    void HTS_Engine_set_speed(HTS_Engine * engine, double f)
    void HTS_Engine_add_half_tone(HTS_Engine * engine, double f)


# distutils: language = c++

cdef extern from "mecab.h":
    cdef cppclass Mecab:
        char **feature
        int size
        void *model
        void *tagger
        void *lattice

    cdef int Mecab_initialize(Mecab *m)
    cdef int Mecab_load(Mecab *m, const char *dicdir)
    cdef int Mecab_analysis(Mecab *m, const char *str)
    cdef int Mecab_print(Mecab *m)
    int Mecab_get_size(Mecab *m)
    char **Mecab_get_feature(Mecab *m)
    cdef int Mecab_refresh(Mecab *m)
    cdef int Mecab_clear(Mecab *m)



# distutils: language = c++

from libc.stdio cimport FILE

cdef extern from "njd.h":
    cdef cppclass NJDNode:
        char *string
        char *pos
        char *pos_group1
        char *pos_group2
        char *pos_group3
        char *ctype
        char *cform
        char *orig
        char *read
        char *pron
        int acc
        int mora_size
        char *chain_rule
        int chain_flag
        NJDNode *prev
        NJDNode *next

    void NJDNode_initialize(NJDNode * node)
    void NJDNode_set_string(NJDNode * node, const char *str)
    void NJDNode_set_pos(NJDNode * node, const char *str)
    void NJDNode_set_pos_group1(NJDNode * node, const char *str)
    void NJDNode_set_pos_group2(NJDNode * node, const char *str)
    void NJDNode_set_pos_group3(NJDNode * node, const char *str)
    void NJDNode_set_ctype(NJDNode * node, const char *str)
    void NJDNode_set_cform(NJDNode * node, const char *str)
    void NJDNode_set_orig(NJDNode * node, const char *str)
    void NJDNode_set_read(NJDNode * node, const char *str)
    void NJDNode_set_pron(NJDNode * node, const char *str)
    void NJDNode_set_acc(NJDNode * node, int acc)
    void NJDNode_set_mora_size(NJDNode * node, int size)
    void NJDNode_set_chain_rule(NJDNode * node, const char *str)
    void NJDNode_set_chain_flag(NJDNode * node, int flag)
    void NJDNode_add_read(NJDNode * node, const char *str)
    void NJDNode_add_pron(NJDNode * node, const char *str)
    void NJDNode_add_acc(NJDNode * node, int acc)
    void NJDNode_add_mora_size(NJDNode * node, int size)
    const char *NJDNode_get_string(NJDNode * node)
    const char *NJDNode_get_pos(NJDNode * node)
    const char *NJDNode_get_pos_group1(NJDNode * node)
    const char *NJDNode_get_pos_group2(NJDNode * node)
    const char *NJDNode_get_pos_group3(NJDNode * node)
    const char *NJDNode_get_ctype(NJDNode * node)
    const char *NJDNode_get_cform(NJDNode * node)
    const char *NJDNode_get_orig(NJDNode * node)
    const char *NJDNode_get_read(NJDNode * node)
    const char *NJDNode_get_pron(NJDNode * node)
    int NJDNode_get_acc(NJDNode * node)
    int NJDNode_get_mora_size(NJDNode * node)
    const char *NJDNode_get_chain_rule(NJDNode * node)
    int NJDNode_get_chain_flag(NJDNode * node)
    void NJDNode_load(NJDNode * node, const char *str)
    NJDNode *NJDNode_insert(NJDNode * prev, NJDNode * next, NJDNode * node)
    void NJDNode_copy(NJDNode * node1, NJDNode * node2)
    void NJDNode_print(NJDNode * node)
    void NJDNode_fprint(NJDNode * node, FILE * fp)
    void NJDNode_sprint(NJDNode * node, char *buff, const char *split_code)
    void NJDNode_clear(NJDNode * node)

    cdef cppclass NJD:
        NJDNode *head
        NJDNode *tail

    void NJD_initialize(NJD * njd)
    void NJD_load(NJD * njd, const char *str)
    void NJD_load_from_fp(NJD * njd, FILE * fp)
    int NJD_get_size(NJD * njd)
    void NJD_push_node(NJD * njd, NJDNode * node)
    void NJD_remove_silent_node(NJD * njd)
    void NJD_print(NJD * njd)
    void NJD_fprint(NJD * njd, FILE * fp)
    void NJD_sprint(NJD * njd, char *buff, const char *split_code)
    void NJD_refresh(NJD * njd)
    void NJD_clear(NJD * wl)

cdef extern from "njd_set_accent_phrase.h":
    void njd_set_accent_phrase(NJD * njd)

cdef extern from "njd_set_accent_type.h":
    void njd_set_accent_type(NJD * njd)

cdef extern from "njd_set_digit.h":
    void njd_set_digit(NJD * njd)

cdef extern from "njd_set_long_vowel.h":
    void njd_set_long_vowel(NJD * njd)

cdef extern from "njd_set_pronunciation.h":
    void njd_set_pronunciation(NJD * njd)

cdef extern from "njd_set_unvoiced_vowel.h":
    void njd_set_unvoiced_vowel(NJD * njd)



# coding: utf-8
# cython: boundscheck=True, wraparound=True
# cython: c_string_type=unicode, c_string_encoding=ascii

import numpy as np

cimport numpy as np
np.import_array()

cimport cython
from libc.stdlib cimport malloc, free

from htsengine cimport HTS_Engine
from htsengine cimport (
    HTS_Engine_initialize, HTS_Engine_load, HTS_Engine_clear, HTS_Engine_refresh,
    HTS_Engine_get_sampling_frequency, HTS_Engine_get_fperiod,
    HTS_Engine_set_speed, HTS_Engine_add_half_tone,
    HTS_Engine_synthesize_from_strings,
    HTS_Engine_get_generated_speech, HTS_Engine_get_nsamples
)

cdef class HTSEngine(object):
    """HTSEngine

    Args:
        voice (bytes): File path of htsvoice.
    """
    cdef HTS_Engine* engine

    def __cinit__(self, bytes voice=b"htsvoice/mei_normal.htsvoice"):
        self.engine = new HTS_Engine()

        HTS_Engine_initialize(self.engine)

        if self.load(voice) != 1:
          self.clear()
          raise RuntimeError("Failed to initalize HTS_Engine")

    def load(self, bytes voice):
        cdef char* voices = voice
        cdef char ret
        ret = HTS_Engine_load(self.engine, &voices, 1)
        return ret

    def get_sampling_frequency(self):
        """Get sampling frequency
        """
        return HTS_Engine_get_sampling_frequency(self.engine)

    def get_fperiod(self):
        """Get frame period"""
        return HTS_Engine_get_fperiod(self.engine)

    def set_speed(self, speed=1.0):
        """Set speed

        Args:
            speed (float): speed
        """
        HTS_Engine_set_speed(self.engine, speed)

    def add_half_tone(self, half_tone=0.0):
        """Additional half tone in log-f0

        Args:
            half_tone (float): additional half tone
        """
        HTS_Engine_add_half_tone(self.engine, half_tone)

    def synthesize(self, list labels):
        """Synthesize waveform from list of full-context labels

        Args:
            labels: full context labels

        Returns:
            np.ndarray: speech waveform
        """
        self.synthesize_from_strings(labels)
        x = self.get_generated_speech()
        self.refresh()
        return x

    def synthesize_from_strings(self, list labels):
        """Synthesize from strings"""
        cdef size_t num_lines = len(labels)
        cdef char **lines = <char**> malloc((num_lines + 1) * sizeof(char*))
        for n in range(len(labels)):
            lines[n] = <char*>labels[n]

        cdef char ret = HTS_Engine_synthesize_from_strings(self.engine, lines, num_lines)
        free(lines)
        if ret != 1:
            raise RuntimeError("Failed to run synthesize_from_strings")

    def get_generated_speech(self):
        """Get generated speech"""
        cdef size_t nsamples = HTS_Engine_get_nsamples(self.engine)
        cdef np.ndarray speech = np.zeros([nsamples], dtype=np.float64)
        cdef size_t index
        for index in range(nsamples):
            speech[index] = HTS_Engine_get_generated_speech(self.engine, index)
        return speech

    def get_fullcontext_label_format(self):
        """Get full-context label format"""
        return (<bytes>HTS_Engine_get_fullcontext_label_format(self.engine)).decode("utf-8")

    def refresh(self):
        HTS_Engine_refresh(self.engine)

    def clear(self):
        HTS_Engine_clear(self.engine)

    def __dealloc__(self):
        self.clear()
        del self.engine



# coding: utf-8
# cython: boundscheck=True, wraparound=True
# cython: c_string_type=unicode, c_string_encoding=ascii

import numpy as np

cimport numpy as np
np.import_array()

cimport cython

from openjtalk.mecab cimport Mecab, Mecab_initialize, Mecab_load, Mecab_analysis
from openjtalk.mecab cimport Mecab_get_feature, Mecab_get_size, Mecab_refresh, Mecab_clear
from openjtalk.njd cimport NJD, NJD_initialize, NJD_refresh, NJD_print, NJD_clear
from openjtalk cimport njd as _njd
from openjtalk.jpcommon cimport JPCommon, JPCommon_initialize,JPCommon_make_label
from openjtalk.jpcommon cimport JPCommon_get_label_size, JPCommon_get_label_feature
from openjtalk.jpcommon cimport JPCommon_refresh, JPCommon_clear
from openjtalk cimport njd2jpcommon
from openjtalk.text2mecab cimport text2mecab
from openjtalk.mecab2njd cimport mecab2njd
from openjtalk.njd2jpcommon cimport njd2jpcommon

cdef njd_node_get_string(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_string(node))).decode("utf-8")

cdef njd_node_get_pos(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_pos(node))).decode("utf-8")

cdef njd_node_get_pos_group1(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_pos_group1(node))).decode("utf-8")

cdef njd_node_get_pos_group2(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_pos_group2(node))).decode("utf-8")

cdef njd_node_get_pos_group3(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_pos_group3(node))).decode("utf-8")

cdef njd_node_get_ctype(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_ctype(node))).decode("utf-8")

cdef njd_node_get_cform(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_cform(node))).decode("utf-8")

cdef njd_node_get_orig(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_orig(node))).decode("utf-8")

cdef njd_node_get_read(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_read(node))).decode("utf-8")

cdef njd_node_get_pron(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_pron(node))).decode("utf-8")

cdef njd_node_get_acc(_njd.NJDNode* node):
    return _njd.NJDNode_get_acc(node)

cdef njd_node_get_mora_size(_njd.NJDNode* node):
    return _njd.NJDNode_get_mora_size(node)

cdef njd_node_get_chain_rule(_njd.NJDNode* node):
    return (<bytes>(_njd.NJDNode_get_chain_rule(node))).decode("utf-8")

cdef njd_node_get_chain_flag(_njd.NJDNode* node):
      return _njd.NJDNode_get_chain_flag(node)


cdef njd_node_print(_njd.NJDNode* node):
  return "{},{},{},{},{},{},{},{},{},{},{}/{},{},{}".format(
    njd_node_get_string(node),
    njd_node_get_pos(node),
    njd_node_get_pos_group1(node),
    njd_node_get_pos_group2(node),
    njd_node_get_pos_group3(node),
    njd_node_get_ctype(node),
    njd_node_get_cform(node),
    njd_node_get_orig(node),
    njd_node_get_read(node),
    njd_node_get_pron(node),
    njd_node_get_acc(node),
    njd_node_get_mora_size(node),
    njd_node_get_chain_rule(node),
    njd_node_get_chain_flag(node)
    )


cdef njd_print(_njd.NJD* njd):
    cdef _njd.NJDNode* node = njd.head
    njd_results = []
    while node is not NULL:
      njd_results.append(njd_node_print(node))
      node = node.next
    return njd_results

cdef class OpenJTalk(object):
    """OpenJTalk

    Args:
        dn_mecab (bytes): Dictionaly path for MeCab.
    """
    cdef Mecab* mecab
    cdef NJD* njd
    cdef JPCommon* jpcommon

    def __cinit__(self, bytes dn_mecab=b"/usr/local/dic"):
        self.mecab = new Mecab()
        self.njd = new NJD()
        self.jpcommon = new JPCommon()

        Mecab_initialize(self.mecab)
        NJD_initialize(self.njd)
        JPCommon_initialize(self.jpcommon)

        r = self._load(dn_mecab)
        if r != 1:
          self._clear()
          raise RuntimeError("Failed to initalize Mecab")


    def _clear(self):
      Mecab_clear(self.mecab)
      NJD_clear(self.njd)
      JPCommon_clear(self.jpcommon)

    def _load(self, bytes dn_mecab):
        return Mecab_load(self.mecab, dn_mecab)


    def run_frontend(self, text, verbose=0):
        """Run OpenJTalk's text processing frontend
        """
        if isinstance(text, str):
          text = text.encode("utf-8")
        cdef char buff[8192]
        text2mecab(buff, text)
        Mecab_analysis(self.mecab, buff)
        mecab2njd(self.njd, Mecab_get_feature(self.mecab), Mecab_get_size(self.mecab))
        _njd.njd_set_pronunciation(self.njd)
        _njd.njd_set_digit(self.njd)
        _njd.njd_set_accent_phrase(self.njd)
        _njd.njd_set_accent_type(self.njd)
        _njd.njd_set_unvoiced_vowel(self.njd)
        _njd.njd_set_long_vowel(self.njd)
        njd2jpcommon(self.jpcommon, self.njd)
        JPCommon_make_label(self.jpcommon)

        cdef int label_size = JPCommon_get_label_size(self.jpcommon)
        cdef char** label_feature
        label_feature = JPCommon_get_label_feature(self.jpcommon)

        labels = []
        for i in range(label_size):
          # This will create a copy of c string
          # http://cython.readthedocs.io/en/latest/src/tutorial/strings.html
          labels.append(<unicode>label_feature[i])

        njd_results = njd_print(self.njd)

        if verbose > 0:
          NJD_print(self.njd)

        # Note that this will release memory for label feature
        JPCommon_refresh(self.jpcommon)
        NJD_refresh(self.njd)
        Mecab_refresh(self.mecab)

        return njd_results, labels

    def g2p(self, text, kana=False, join=True):
        """Grapheme-to-phoeneme (G2P) conversion
        """
        njd_results, labels = self.run_frontend(text)
        if not kana:
            prons = list(map(lambda s: s.split("-")[1].split("+")[0], labels[1:-1]))
            if join:
                prons = " ".join(prons)
            return prons

        # kana
        prons = []
        for n in njd_results:
            row = n.split(",")
            if row[1] == "記号":
                p = row[0]
            else:
                p = row[9]
            # remove special chars
            for c in "’":
                p = p.replace(c,"")
            prons.append(p)
        if join:
            prons = "".join(prons)
        return prons

    def __dealloc__(self):
        self._clear()
        del self.mecab
        del self.njd
        del self.jpcommon



import os
import subprocess
from distutils.version import LooseVersion
from glob import glob
from os.path import exists, join
from subprocess import run

import numpy as np
import setuptools.command.build_py
import setuptools.command.develop
from setuptools import Extension, find_packages, setup

version = "0.1.0"

min_cython_ver = "0.21.0"
try:
    import Cython

    ver = Cython.__version__
    _CYTHON_INSTALLED = ver >= LooseVersion(min_cython_ver)
except ImportError:
    _CYTHON_INSTALLED = False

try:
    if not _CYTHON_INSTALLED:
        raise ImportError("No supported version of Cython installed.")
    from Cython.Distutils import build_ext

    cython = True
except ImportError:
    cython = False

if cython:
    ext = ".pyx"
    cmdclass = {"build_ext": build_ext}
else:
    ext = ".cpp"
    cmdclass = {}
    if not os.path.exists(join("pyopenjtalk", "openjtalk" + ext)):
        raise RuntimeError("Cython is required to generate C++ code")

# open_jtalk sources
src_top = join("lib", "open_jtalk", "src")

# generate config.h for mecab
# NOTE: need to run cmake to generate config.h
# we could do it on python side but it would be very tricky,
# so far let's use cmake tool
if not exists(join(src_top, "mecab", "src", "config.h")):
    cwd = os.getcwd()
    build_dir = join(src_top, "build")
    os.makedirs(build_dir, exist_ok=True)
    os.chdir(build_dir)
    run(["cmake", ".."])
    os.chdir(cwd)

all_src = []
include_dirs = []
for s in [
    "jpcommon",
    "mecab/src",
    "mecab2njd",
    "njd",
    "njd2jpcommon",
    "njd_set_accent_phrase",
    "njd_set_accent_type",
    "njd_set_digit",
    "njd_set_long_vowel",
    "njd_set_pronunciation",
    "njd_set_unvoiced_vowel",
    "text2mecab",
]:
    all_src += glob(join(src_top, s, "*.c"))
    all_src += glob(join(src_top, s, "*.cpp"))
    include_dirs.append(join(os.getcwd(), src_top, s))

# Extension for OpenJTalk frontend
ext_modules = [
    Extension(
        name="pyopenjtalk.openjtalk",
        sources=[join("pyopenjtalk", "openjtalk" + ext)] + all_src,
        include_dirs=[np.get_include()] + include_dirs,
        extra_compile_args=[],
        extra_link_args=[],
        language="c++",
        define_macros=[
            ("HAVE_CONFIG_H", None),
            ("DIC_VERSION", 102),
            ("MECAB_DEFAULT_RC", '"dummy"'),
            ("PACKAGE", '"open_jtalk"'),
            ("VERSION", '"1.10"'),
            ("CHARSET_UTF_8", None),
        ],
    )
]

# Extension for HTSEngine backend
htsengine_src_top = join("lib", "hts_engine_API", "src")
all_htsengine_src = glob(join(htsengine_src_top, "lib", "*.c"))
ext_modules += [
    Extension(
        name="pyopenjtalk.htsengine",
        sources=[join("pyopenjtalk", "htsengine" + ext)] + all_htsengine_src,
        include_dirs=[np.get_include(), join(htsengine_src_top, "include")],
        extra_compile_args=[],
        extra_link_args=[],
        language="c++",
    )
]

# Adapted from https://github.com/pytorch/pytorch
cwd = os.path.dirname(os.path.abspath(__file__))
if os.getenv("PYOPENJTALK_BUILD_VERSION"):
    version = os.getenv("PYOPENJTALK_BUILD_VERSION")
else:
    try:
        sha = (
            subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd)
            .decode("ascii")
            .strip()
        )
        version += "+" + sha[:7]
    except subprocess.CalledProcessError:
        pass
    except IOError:  # FileNotFoundError for python 3
        pass


class build_py(setuptools.command.build_py.build_py):
    def run(self):
        self.create_version_file()
        setuptools.command.build_py.build_py.run(self)

    @staticmethod
    def create_version_file():
        global version, cwd
        print("-- Building version " + version)
        version_path = os.path.join(cwd, "pyopenjtalk", "version.py")
        with open(version_path, "w") as f:
            f.write("__version__ = '{}'\n".format(version))


class develop(setuptools.command.develop.develop):
    def run(self):
        build_py.create_version_file()
        setuptools.command.develop.develop.run(self)


cmdclass["build_py"] = build_py
cmdclass["develop"] = develop


with open("README.md", "r") as fd:
    long_description = fd.read()

setup(
    name="pyopenjtalk",
    version=version,
    description="A python wrapper for OpenJTalk",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ryuichi Yamamoto",
    author_email="zryuichi@gmail.com",
    url="https://github.com/r9y9/pyopenjtalk",
    license="MIT",
    packages=find_packages(),
    package_data={"": ["htsvoice/*"]},
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    install_requires=[
        "numpy >= 1.8.0",
        "cython >= " + min_cython_ver,
        "six",
    ],
    tests_require=["nose", "coverage"],
    extras_require={
        "docs": ["sphinx_rtd_theme", "nbsphinx", "pandoc", "ipython", "jupyter"],
        "test": ["pytest", "scipy"],
    },
    classifiers=[
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
    ],
    keywords=["OpenJTalk", "Research"],
)



# distutils: language = c++

from libc.stdio cimport FILE

cdef extern from "jpcommon.h":
    cdef cppclass JPCommonNode:
       char *pron
       char *pos
       char *ctype
       char *cform
       int acc
       int chain_flag
       void *prev
       void *next

    cdef cppclass JPCommon:
       JPCommonNode *head
       JPCommonNode *tail
       void *label

    void JPCommon_initialize(JPCommon * jpcommon)
    void JPCommon_push(JPCommon * jpcommon, JPCommonNode * node)
    void JPCommon_make_label(JPCommon * jpcommon)
    int JPCommon_get_label_size(JPCommon * jpcommon)
    char **JPCommon_get_label_feature(JPCommon * jpcommon)
    void JPCommon_print(JPCommon * jpcommon)
    void JPCommon_fprint(JPCommon * jpcommon, FILE * fp)
    void JPCommon_refresh(JPCommon * jpcommon)
    void JPCommon_clear(JPCommon * jpcommon)



import os
from os.path import exists

import pkg_resources
import six

if six.PY2:
    from urllib import urlretrieve
else:
    from urllib.request import urlretrieve

import tarfile

try:
    from .version import __version__  # NOQA
except ImportError:
    raise ImportError("BUG: version.py doesn't exist. Please file a bug report.")

from .htsengine import HTSEngine
from .openjtalk import OpenJTalk

# Dictionary directory
# defaults to the package directory where the dictionary will be automatically downloaded
OPEN_JTALK_DICT_DIR = os.environ.get(
    "OPEN_JTALK_DICT_DIR",
    pkg_resources.resource_filename(__name__, "open_jtalk_dic_utf_8-1.11"),
).encode("ascii")
_DICT_URL = (
    "https://downloads.sourceforge.net/open-jtalk/open_jtalk_dic_utf_8-1.11.tar.gz"
)

# Default mei_normal.voice for HMM-based TTS
DEFAULT_HTS_VOICE = pkg_resources.resource_filename(
    __name__, "htsvoice/mei_normal.htsvoice"
).encode("ascii")

# Global instance of OpenJTalk
_global_jtalk = None
# Global instance of HTSEngine
# mei_normal.voice is used as default
_global_htsengine = None


def _extract_dic():
    global OPEN_JTALK_DICT_DIR
    filename = pkg_resources.resource_filename(__name__, "dic.tar.gz")
    print('Downloading: "{}"'.format(_DICT_URL))
    urlretrieve(_DICT_URL, filename)
    print("Extracting tar file {}".format(filename))
    with tarfile.open(filename, mode="r|gz") as f:
        f.extractall(path=pkg_resources.resource_filename(__name__, ""))
    OPEN_JTALK_DICT_DIR = pkg_resources.resource_filename(
        __name__, "open_jtalk_dic_utf_8-1.11"
    ).encode("ascii")
    os.remove(filename)


def _lazy_init():
    if not exists(OPEN_JTALK_DICT_DIR):
        _extract_dic()


def g2p(*args, **kwargs):
    """Grapheme-to-phoeneme (G2P) conversion

    This is just a convenient wrapper around `run_frontend`.

    Args:
        text (str): Unicode Japanese text.
        kana (bool): If True, returns the pronunciation in katakana, otherwise in phone.
          Default is False.
        join (bool): If True, concatenate phones or katakana's into a single string.
          Default is True.

    Returns:
        str or list: G2P result in 1) str if join is True 2) list if join is False.
    """
    global _global_jtalk
    if _global_jtalk is None:
        _lazy_init()
        _global_jtalk = OpenJTalk(dn_mecab=OPEN_JTALK_DICT_DIR)
    return _global_jtalk.g2p(*args, **kwargs)


def extract_fullcontext(text):
    """Extract full-context labels from text

    Args:
        text (str): Input text

    Returns:
        list: List of full-context labels
    """
    # note: drop first return
    _, labels = run_frontend(text)
    return labels


def synthesize(labels, speed=1.0, half_tone=0.0):
    """Run OpenJTalk's speech synthesis backend

    Args:
        labels (list): Full-context labels
        speed (float): speech speed rate. Default is 1.0.
        half_tone (float): additional half-tone. Default is 0.

    Returns:
        np.ndarray: speech waveform (dtype: np.float64)
        int: sampling frequency (defualt: 48000)
    """
    if isinstance(labels, tuple) and len(labels) == 2:
        labels = labels[1]

    global _global_htsengine
    if _global_htsengine is None:
        _global_htsengine = HTSEngine(DEFAULT_HTS_VOICE)
    sr = _global_htsengine.get_sampling_frequency()
    _global_htsengine.set_speed(speed)
    _global_htsengine.add_half_tone(half_tone)
    return _global_htsengine.synthesize(labels), sr


def tts(text, speed=1.0, half_tone=0.0):
    """Text-to-speech

    Args:
        text (str): Input text
        speed (float): speech speed rate. Default is 1.0.
        half_tone (float): additional half-tone. Default is 0.

    Returns:
        np.ndarray: speech waveform (dtype: np.float64)
        int: sampling frequency (defualt: 48000)
    """
    return synthesize(extract_fullcontext(text), speed, half_tone)


def run_frontend(text, verbose=0):
    """Run OpenJTalk's text processing frontend

    Args:
        text (str): Unicode Japanese text.
        verbose (int): Verbosity. Default is 0.

    Returns:
        tuple: Pair of 1) NJD_print and 2) JPCommon_make_label.
        The latter is the full-context labels in HTS-style format.
    """
    global _global_jtalk
    if _global_jtalk is None:
        _lazy_init()
        _global_jtalk = OpenJTalk(dn_mecab=OPEN_JTALK_DICT_DIR)
    return _global_jtalk.run_frontend(text, verbose)
